@@ -1,5 +1,5 @@
 from .file import File
-from .graph import GraphNode, NodeArg, NodeCall, Out, OutDim
+from .graph import GraphNode, NodeArg, NodeCall, Out
 from .node import Node, NodeLet, NodeLetType, NodeType, TypeConverter, With, WithSet
 from .use import Use, UseByLocal, UseByAuthor, UseByWeb
 from .variable import Expr, Operator, OperatorFn, Shape, Variable, clone_value, is_estimable, is_hint
