@@ -16,13 +16,6 @@ class Out:
         return f'{name}${id}'
 
 
-class OutDim:
-    def __init__(self, out, dim):
-        super().__init__()
-        self.out = out
-        self.dim = dim
-
-
 class NodeArg:
     def __init__(self, name, value):
         super().__init__()
@@ -34,11 +27,12 @@ class NodeArg:
 
 
 class NodeCall:
-    def __init__(self, name, inputs, args):
+    def __init__(self, name, inputs, args, repeat):
         super().__init__()
         self.name = name
         self.inputs = inputs
         self.args = args
+        self.repeat = repeat
 
     def get_arg(self, name):
         assert self.args is not None, f'unexpected empty args: {self.name}'
@@ -65,7 +59,12 @@ class NodeCall:
         else:
             args = ''
 
-        return self.name + inputs + args
+        if self.repeat is not None:
+            repeat = f' * {self.repeat}'
+        else:
+            repeat = ''
+
+        return self.name + inputs + args + repeat
 
 
 class GraphNode:
