@@ -36,3 +36,6 @@ class DataNode(ExternNodeBase, metaclass=abc.ABCMeta):
     @abc.abstractmethod
     def get_eval_dataset(self):
         raise NotImplementedError
+
+    def to(self, device):
+        return self
