@@ -1,5 +1,6 @@
 import abc
 import os
+import pickle
 
 import torch
 import inflection
@@ -7,24 +8,57 @@ import inflection
 from .data import DataNode
 from .node import Node
 from .optim import OptimNode
+from ... import device
 from ...writer import ExecWriter
 
 
 class ExecNode(metaclass=abc.ABCMeta):
-    def __init__(self, env, root, **kwargs):
+    _devices: list
+    _writer: ExecWriter
+
+    def __init__(self, root, **kwargs):
         from ..node import NodeIR
 
+        self._nodes = []
         for k, v in kwargs.items():
             if isinstance(v, NodeIR):
-                v = v.build(root)
+                self._nodes.append(k)
             setattr(self, k, v)
-        self._writer = ExecWriter(env, self)
+
+        # dump
+        with open('dump.pkl', 'wb') as f:
+            pickle.dump(self, f)
+
+        self._initialize(root)
+
+    def _initialize(self, root):
+        with root.use_exec(self):
+            for k, v in self.get_nodes().items():
+                setattr(self, k, v.build(root.node_root))
+
+        self._devices = device.parse_device(root.env.devices)
+        self._writer = ExecWriter(root.env, self)
+
+    @classmethod
+    def load(cls, root, f):
+        exec = pickle.load(f)
+        exec._initialize(root)
+        return exec
 
     @classmethod
     def link_node(cls, last, new):
-        from ..node import merge_shapes
+        """link the shapes between nodes, such as data->model"""
+        from ..node import link_shapes
 
-        merge_shapes(last.get_output_shapes(), new.get_input_shapes())
+        link_shapes(last.get_output_shapes(), new.get_input_shapes())
+
+    def get_nodes(self):
+        return {k: getattr(self, k) for k in self._nodes}
+
+    def to(self, node):
+        # TODO: https://github.com/pytorch/pytorch/blob/master/torch/distributed/launch.py
+        # print(self._devices)
+        return node.to('cuda:0')
 
     def get_name(self):
         return self.__class__.__name__
@@ -40,9 +74,9 @@ class Trainer(ExecNode, metaclass=abc.ABCMeta):
     epoch: int
     batch_size: int
 
-    def __init__(self, env, root, data, model, **kwargs):
+    def __init__(self, root, data, model, **kwargs):
         self.__class__.link_node(data, model)
-        super().__init__(env, root, data=data, model=model, **kwargs)
+        super().__init__(root, data=data, model=model, **kwargs)
 
     def train(self):
         # Step 1. ready to train
@@ -78,12 +112,14 @@ class Trainer(ExecNode, metaclass=abc.ABCMeta):
 
     def _train_begin(self):
         self.optimizer.initialize(self.model)
+        for name, node in self.get_nodes().items():
+            setattr(self, name, self.to(node))
 
     def _train_epoch_begin(self, epoch, metrics):
         self.model.train()
 
     def _train_iter_begin(self, data):
-        return {'x': data[0]}, {'y': data[1]}
+        return {'x': self.to(data[0])}, {'y': self.to(data[1])}
 
     def _train_iter_end(self, metrics, x, y, y_pred, loss):
         metrics['loss'] += loss.item()
