@@ -25,3 +25,6 @@ class OptimNode(ExternNodeBase, metaclass=abc.ABCMeta):
 
     def step(self):
         self._inner.step()
+
+    def to(self, device):
+        return self
