@@ -1,7 +1,9 @@
-from copy import copy
+from copy import copy, deepcopy
 import importlib.util
+import sys
 
 import torch.nn as nn
+from inflection import underscore
 
 from .. import ast, std
 from ..parser import Parser
@@ -13,7 +15,10 @@ from .variable import build as build_variable
 
 
 _BuiltinsInput = ['AssertShape', ]
-_BuiltinsNode = ['Transform', 'Concat', ]
+_BuiltinsNode = [
+    'Transform', 'ToLinear',
+    'Concat',
+]
 
 
 def make_graph(id, variables):
@@ -43,6 +48,8 @@ def _assert_tensor_graph_name(node: ast.GraphNode, expects,
         assert isinstance(
             call.inputs, inputs), f'Expected {inputs.__name__} inputs, Given {type_name}'
 
+    assert call.repeat is None, f'unexpected repeat syntax: {call.name}'
+
     if not args:
         assert call.args is None, f'Expected {expects} node, Given {repr(call)}'
     if sized:
@@ -50,7 +57,7 @@ def _assert_tensor_graph_name(node: ast.GraphNode, expects,
     return call
 
 
-def merge_shapes(last_outputs, new_inputs):
+def link_shapes(last_outputs, new_inputs):
     for name in last_outputs.keys():
         last_output = last_outputs[name]
 
@@ -72,7 +79,7 @@ def merge_shapes(last_outputs, new_inputs):
             if not ast.is_hint(last_dim):
                 # replace
                 if ast.is_hint(new_dim):
-                    new_dim.ty = ast.NodeLetType.INT
+                    new_dim = new_dim.get_hint()
                     new_dim.value = last_dim
                 # test value
                 else:
@@ -90,6 +97,7 @@ def _assert_equal(last, new):
         # TODO: FUTURE: implement comparing hinted values
         print('warning: comparing hinted values is not supported yet!')
     else:
+        last, new = int(last), int(new)
         assert last == new, f'Expected {last}, Given {new}'
 
 
@@ -103,6 +111,7 @@ class NodeIR:
         self.id = None
         self.input = None
         self.output = None
+        self.repeat = None
 
     def clone(self, root, variables):
         graph = self.graph.clone(root.generate_id(), variables)
@@ -110,8 +119,9 @@ class NodeIR:
 
         cloned = NodeIR(self.name, graph, tensor_graph)
         cloned.id = self.id
-        cloned.input = copy(self.input)
-        cloned.output = copy(self.output)
+        cloned.input = deepcopy(self.input)
+        cloned.output = deepcopy(self.output)
+        cloned.repeat = ast.clone_value(self.repeat, variables)
         return cloned
 
     def get_input_shapes(self):
@@ -139,11 +149,56 @@ class NodeIR:
                                for k in self.get_input_shapes().keys()}
         output = self.output or {k: ast.Out(1, k)
                                  for k in self.get_output_shapes().keys()}
+
+        if self.repeat is not None:
+            repeat = build_variable(self.repeat)
+
+            # remove nodes
+            if repeat == 0:
+                # the input-output shapes should be same
+                link_shapes(self.get_input_shapes(),
+                            self.get_output_shapes())
+                # clean up graph
+                self.tensor_graph = []
+
+            # repeat nodes
+            elif repeat > 1:
+                base_graph = copy(self.tensor_graph)
+
+                def update_out_id(out):
+                    for node in reversed(base_graph):
+                        shapes = node.get_output_shapes()
+                        if out.name in shapes.keys():
+                            out.id = node.id
+                            return
+                    raise Exception('unreachable code')
+
+                for _ in range(repeat-1):
+                    for node in base_graph:
+                        node = node.clone(root, [])
+                        dims = node.graph.unload_dims()
+
+                        last_outputs = self.get_output_shapes()
+                        new_inputs = node.get_input_shapes()
+                        link_shapes(last_outputs, new_inputs)
+
+                        for out in node.input.values():
+                            update_out_id(out)
+                        for out in node.output.values():
+                            update_out_id(out)
+
+                        node.graph.load_dims_weakly(dims)
+                        self.tensor_graph.append(node)
+
+            elif repeat != 1:
+                raise Exception('unestimable repeat value')
+
         tensor_graph = [n.build(root) for n in self.tensor_graph]
         if len(tensor_graph) == 1:
             node = tensor_graph[0]
             if isinstance(node, (DataNode, OptimNode)):
                 return tensor_graph[0]
+
         return NodeExecutable(self.name, input, output, tensor_graph)
 
     def __repr__(self):
@@ -224,6 +279,7 @@ class NodeContainer:
                 # Step 1. get the node
                 callee = self.get(call.name)
                 callee.id = node.id
+                callee.repeat = self.graph.replace(call.repeat)
 
                 # Step 2. apply variables
                 if call.args:
@@ -252,7 +308,7 @@ class NodeContainer:
                                     for k, x in inputs.items()}
                     new_inputs = callee.get_input_shapes()
 
-                    merge_shapes(last_outputs, new_inputs)
+                    link_shapes(last_outputs, new_inputs)
 
                     # identity
                     new_outputs = callee.get_output_shapes()
@@ -269,7 +325,7 @@ class NodeContainer:
             # Step 6. merge dedicated shapes
             if node.shapes:
                 outputs = self.get_output_shapes()
-                merge_shapes(node.shapes, outputs)
+                link_shapes(node.shapes, outputs)
 
             # Step 7. store id
             self.last_id = node.id
@@ -310,6 +366,13 @@ class NodeContainer:
             return self.children[name].clone(self.root, [])
         return self.parent.get(name)
 
+    def get_variable(self, name, candidates=set()):
+        variables = self.graph.get_shortcuts()
+        if name in variables:
+            return variables[name]
+        candidates.update(*variables)
+        return self.parent.get_variable(name, candidates)
+
     def build(self):
         return NodeIR(self.name, self.graph, self.tensor_graph)
 
@@ -317,33 +380,36 @@ class NodeContainer:
         return self.parent._build(file, parent)
 
     def _build_builtins(self, name, node):
-        if name == 'Transform':
-            return self._build_transform(node)
-        if name == 'Concat':
-            return self._build_concat(node)
-        raise NotImplementedError
+        return getattr(self, f'_build_{underscore(name)}')(name, node)
 
-    def _build_transform(self, node):
-        call = _assert_tensor_graph_name(node, ['Transform'], sized=True)
+    def _build_transform(self, name, node, linear=False):
+        sized = not linear
+        call = _assert_tensor_graph_name(node, [name], sized=sized)
 
         # Step 1. get the IO
         inputs = self.get_output_shapes()
-        outputs = node.shapes
+        outputs = node.shapes or inputs
+
+        if linear:
+            outputs = {k: ast.Shape([v.product()], k)
+                       for k, v in outputs.items()}
 
         # Step 2. match the tuple
         input_keys = inputs.keys()
-        output_keys = outputs.keys()
-        if input_keys != output_keys:
-            input_keys = ', '.join(input_keys)
-            output_keys = ', '.join(output_keys)
-            raise Exception(
-                f'Expected tuple ({input_keys}), Given ({output_keys})')
+        if sized:
+            output_keys = outputs.keys()
+            if input_keys != output_keys:
+                input_keys = ', '.join(input_keys)
+                output_keys = ', '.join(output_keys)
+                raise Exception(
+                    f'Expected tuple ({input_keys}), Given ({output_keys})')
 
         # Step 3. match the size
-        for name in input_keys:
-            input = build_variable(inputs[name].product())
-            output = build_variable(outputs[name].product())
-            _assert_equal(input, output)
+        if sized:
+            for name in input_keys:
+                input = build_variable(inputs[name].product())
+                output = build_variable(outputs[name].product())
+                _assert_equal(input, output)
 
         # Step 4. apply to tensor graph
         id = self.last_id + 1
@@ -354,9 +420,12 @@ class NodeContainer:
 
         # Step 5. store
         kwargs = {'output_shapes': outputs}
-        return TensorGraph(node.id, call.name, kwargs, input, output, inputs, outputs)
+        return TensorGraph(node.id, 'Transform', kwargs, input, output, inputs, outputs)
 
-    def _build_concat(self, node):
+    def _build_to_linear(self, name, node):
+        return self._build_transform(name, node, linear=True)
+
+    def _build_concat(self, name, node):
         call = _assert_tensor_graph_name(
             node, ['Concat'], inputs=list, args=True)
 
@@ -431,6 +500,13 @@ class ExternNodeContainer:
         else:
             self.output = node.shapes
 
+    def get_variable(self, name, candidates=set()):
+        variables = self.graph.get_shortcuts()
+        if name in variables:
+            return variables[name]
+        candidates.update(*variables)
+        return self.parent.get_variable(name, candidates)
+
     def build(self):
         # set default IO to "x"
         if not (self.data or self.optim):
@@ -493,53 +569,21 @@ class ExecIR(EnvVars):
         # Step 1. assert each variable is not none
         self._assert_not_none(self._name)
 
-        # Step 2. apply to environment variables
-        with self._use_exec(root):
-            # Step 3. build nodes
-            for var in self._graph.get_variables().values():
-                value = getattr(self, var.name.replace(' ', '_'))
-                if var.ty.is_node():
-                    var.value = root.node_root.get(value)
-                else:
-                    var.value = build_variable(value)
+        # Step 2. build nodes
+        for var in self._graph.get_variables().values():
+            value = getattr(self, var.name.replace(' ', '_'))
+            if var.ty.is_node():
+                var.value = root.node_root.get(value)
+            else:
+                var.value = build_variable(value)
 
-            # Step 3. get variables
-            kwargs = {k: v.value for k,
-                      v in self._graph.get_variables().items()}
+        # Step 3. get variables
+        kwargs = {k.replace(' ', '_'): v.value for k,
+                  v in self._graph.get_variables().items()}
 
-            # Step 4. build to extern exec node
-            node_cls = root.node_root.get_extern(self._name)
-            return node_cls(env=root.env, root=root.node_root, **kwargs)
-
-    def _use_exec(self, root):
-        class ReadOnlyExec:
-            def __init__(self, exec, env):
-                self._exec = exec
-                self._env = env
-
-            def __getattr__(self, name):
-                if name == 'env':
-                    return self._env
-                if not name.startswith('_'):
-                    return getattr(self._exec, name)
-
-                # default behaviour
-                return self.__getattribute__(name)
-
-        class UseExec:
-            def __init__(self, exec, root):
-                super().__init__()
-                self.exec = exec
-                self.root = root
-
-            def __enter__(self):
-                env = self.root.env
-                self.root.node_root.exec = ReadOnlyExec(self.exec, env)
-
-            def __exit__(self, exc_type, exc_val, exc_tb):
-                self.root.node_root.exec = None
-
-        return UseExec(self, root)
+        # Step 4. build to extern exec node
+        node_cls = root.node_root.get_extern(self._name)
+        return node_cls(root=root, **kwargs)
 
 
 class NodeRoot:
@@ -576,6 +620,11 @@ class NodeRoot:
     def add_extern(self, name, path):
         self._externs[name] = path
 
+    def get_variable(self, name, candidates):
+        candidates = ', '.join(candidates)
+        raise Exception(
+            f'no such variable: {name}, candidates are: {candidates}')
+
     def get(self, name):
         if name in self._irs.keys():
             return self._irs[name].clone(self, [])
@@ -597,11 +646,15 @@ class NodeRoot:
         if name in self._externs.keys():
             path = self._externs.pop(name)
 
-            loader = importlib.machinery.SourceFileLoader('n3.extern', path)
+            mod_name = f'n3.extern.{name}'
+            loader = importlib.machinery.SourceFileLoader(mod_name, path)
             spec = importlib.util.spec_from_loader(loader.name, loader)
             mod = importlib.util.module_from_spec(spec)
             loader.exec_module(mod)
 
+            # make sure that pickle can find the module
+            sys.modules[mod_name] = mod
+
             if not hasattr(mod, name):
                 raise Exception(
                     f'node class not defined in python script: {name}, "{path}"')
@@ -632,7 +685,7 @@ class NodeRoot:
             if n.shapes:
                 for x, shapes in n.shapes.items():
                     shapes.dims = container.graph.hint(
-                        ast.Out(id, x), shapes.dims)
+                        container, ast.Out(id, x), shapes.dims)
 
         # Step 4. re-define nodes (with)
         for w in node.withs.values():
@@ -670,7 +723,7 @@ class NodeRoot:
             if n.shapes:
                 for x, shapes in n.shapes.items():
                     shapes.dims = container.graph.hint(
-                        ast.Out(id, x), shapes.dims)
+                        container, ast.Out(id, x), shapes.dims)
 
         # Step 3. make a tensor graph
         num_tensor_graph = len(node.graph)
