@@ -89,18 +89,10 @@ class Graph:
         self._shortcuts = {v.shortcut or v.name: v
                            for v in self._variables.values()}
 
-    def hint(self, out, dims):
+    def hint(self, container, out, dims):
         def _hint(out, dim, value, is_root=False):
             if isinstance(value, ast.Variable):
-                if value.name not in self._shortcuts.keys():
-                    candidates = ', '.join(self._shortcuts.keys())
-                    raise Exception(
-                        f'no such variable: {value.name}, candidates are: {candidates}')
-                output = self._shortcuts[value.name]
-                # hint in-place
-                if output.ty == ast.NodeLetType.DIM and is_root:
-                    output.value = ast.OutDim(out, dim)
-                return output
+                return container.get_variable(value.name)
             if isinstance(value, ast.Expr):
                 value.lhs = _hint(out, dim, value.lhs)
                 value.rhs = _hint(out, dim, value.rhs)
@@ -120,8 +112,31 @@ class Graph:
             value.rhs = self.replace(value.rhs)
         return value
 
+    def unload_dims(self):
+        """pop the dimensional variables' values.
+
+        Returns:
+            dict: the popped values
+        """
+        values = {}
+        for name, var in self._variables.items():
+            if var.ty == ast.NodeLetType.DIM:
+                values[name] = var.value
+                var.value = None
+        return values
+
+    def load_dims_weakly(self, values):
+        """set the dimensional variables' values, when it's not defined yet.
+
+        Args:
+            values (dict): the lastly popped values
+        """
+        for name, value in values.items():
+            var = self._variables[name]
+            if ast.is_hint(var):
+                var.value = value
+
     def apply(self, variables, is_shortcut=False):
-        # TODO: test types
         self_variables = self._shortcuts if is_shortcut else self._variables
         for k, v in variables.items():
             self_variables[k].value = v
@@ -129,6 +144,9 @@ class Graph:
     def assert_estimable(self):
         all(v.assert_estimable() for v in self._variables.values())
 
+    def get_shortcuts(self):
+        return self._shortcuts
+
     def get_variables(self):
         return self._variables
 
