@@ -2,5 +2,5 @@ from n3 import ExternNode
 
 
 class AssertShape(ExternNode):
-    def forward(self, x):
-        return x
+    def forward(self, **kwargs):
+        return kwargs
