@@ -16,12 +16,14 @@ class Conv2D(ExternNode):
 
     def __init__(self, **kwargs):
         super().__init__(**kwargs)
-        self._inner = nn.Conv2d(self.input_channels,
-                                self.output_channels,
-                                kernel_size=self.kernel_size,
-                                stride=self.stride,
-                                padding=self.padding,
-                                bias=self.bias)
+        self._inner = nn.Conv2d(
+            in_channels=self.input_channels,
+            out_channels=self.output_channels,
+            kernel_size=self.kernel_size,
+            stride=self.stride,
+            padding=self.padding,
+            bias=self.bias,
+        )
 
     def forward(self, x):
         return self._inner(x)
